# Copyright 2011-2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from regexlint.indicator_ast import find_offending_line, find_substr_pos


def mark(lineno, d1, d2, text, output_stream):
    output_stream.write("  " + text + "\n")
    output_stream.write("  " + " " * d1 + '^' * (d2-d1) + ' ' + 'here\n')



# Copyright 2011-2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys
import weakref

from regexlint.util import *

from pygments.lexer import RegexLexer, include, using, bygroups
from pygments.token import Other


WHITESPACE = ' \t\n\r\f\v'
DIGITS = '0123456789'
WORD = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ' + DIGITS + '_'

# Special types, others are used in the parser below in normal Pygments
# manner.

PROGRESSION = Other.Progression
ALTERNATION = Other.Alternation

class Node(object):
    def __init__(self, t, start=None, parsed_start=None, data=None):
        self.type = t
        self.data = data # type-dependent
        self.end_data = ''
        self.children = [] # type-dependent

        self.start = start
        self.parsed_start = parsed_start
        # This clause is necessary for simple tokens that don't have children
        # -- they need to essentially be autoclosed, here.
        if start is not None and parsed_start is not None and data is not None:
            self.end = start + len(data)
            self.parsed_end = parsed_start + len(data)
        else:
            self.end = None
            self.parsed_end = None
        self._parent = None
        self._next = None

    def add_child(self, obj):
        obj._parent = weakref.ref(self)
        if self.children:
            self.children[-1]._next = weakref.ref(obj)
        self.children.append(obj)

    def close(self, pos, parsed_pos, data):
        self.end = pos + len(data)
        self.parsed_end = parsed_pos + len(data)
        self.end_data = data

    def next(self):
        if self.children:
            return self.children[0]
        else:
            return self.next_no_children()

    def next_no_children(self):
        # pylint: disable-msg=E1102
        if self._next:
            return self._next()
        p = self._parent
        while p:
            if p()._next:
                return p()._next()
            p = p()._parent
        # pylint: enable-msg=E1102

    def parent(self):
        if self._parent:
            return self._parent()  # pylint: disable-msg=E1102

    def is_descentant_of(self, other):
        # pylint: disable-msg=E1102
        if self is other:
            return True
        p = self._parent
        while p:
            if p() is other:
                return True
            p = p()._parent
        # pylint: enable-msg=E1102

    def reconstruct(self):
        """Return the regex string for this branch of the tree."""
        r = []
        # Special case for repetition
        if self.data and self.data != self.end_data:
            r.append(self.data)
        for i, c in enumerate(self.children):
            r.append(c.reconstruct())
            if self.type is ALTERNATION and i < len(self.children) - 1:
                r.append('|')
        r.append(self.end_data)
        return ''.join(r)

    def __repr__(self):
        return '<%s type=%r data=%r start=%r end=%r %r>' % (
            self.__class__.__name__, self.type, self.data, self.start,
            self.end, self.children)

    def __eq__(self, obj):
        return (self.type == obj.type and self.data == obj.data and
                self.children == obj.children and self.start == obj.start and
                self.end == obj.end)


class RootNode(Node):
    def __init__(self, t, start=None, parsed_start=None, data=None, raw=None, flags=None):
        super(RootNode, self).__init__(t, start, parsed_start, data)
        self.raw = raw
        self.flags = flags


class CharRange(object):
    def __init__(self, a, b):
        self.a = a
        self.b = b
        self.codepoint_a = eval_char(a.data)
        self.codepoint_b = eval_char(b.data)

    def __repr__(self):
        return '<%s %r-%r>' % (self.__class__.__name__, self.a, self.b)


class CharClass(Node):
    def __init__(self, t, start=None, parsed_start=None):
        super(CharClass, self).__init__(t, start, parsed_start)
        self.negated = False

    def close(self, pos, parsed_pos, data):
        super(CharClass, self).close(pos, parsed_pos, data)

        n = []
        it = iter(self.children)
        for child in it:
            c = child.data
            if not n and c == '^':
                # caret is special only when the first char
                self.negated = True
            elif c == '-':
                # dash is special only when it's the first char or directly
                # follows another range (say, [0-9-x]).
                next_child = None
                if n and not isinstance(n[-1], CharRange):
                    try:
                        next_child = it.next()
                    except StopIteration:
                        next_child = None
                if next_child:
                    n.append(CharRange(n.pop(), next_child))
                else:
                    n.append(child)
            else:
                n.append(child)

        # One of the checkers may care about order, so don't set-ify yet
        self.matching_character_codes = []
        for i in n:
            if isinstance(i, CharRange):
                self.matching_character_codes.extend(
                    range(eval_char(i.a.data), eval_char(i.b.data)+1))
            elif i.data == r'\s':
                self.matching_character_codes.extend(map(ord, WHITESPACE))
            elif i.data == r'\w':
                self.matching_character_codes.extend(map(ord, WORD))
            elif i.data == r'\d':
                self.matching_character_codes.extend(map(ord, DIGITS))
            elif i.data == r'\S':
                whitespace = set(map(ord, WHITESPACE))
                self.matching_character_codes.extend(i for i in range(256) if i
                                                     not in whitespace)
            elif i.data == r'\W':
                word = set(map(ord, WORD))
                self.matching_character_codes.extend(i for i in range(256) if i
                                                     not in word)
            elif i.data == r'\D':
                digits = set(map(ord, DIGITS))
                self.matching_character_codes.extend(i for i in range(256) if i
                                                     not in digits)
            else:
                self.matching_character_codes.append(eval_char(i.data))

        self.chars = n

    def reconstruct(self):
        return '[%s%s]' % ('^' if self.negated else '',
                           ''.join((x.a.data + '-' + x.b.data) if
                            isinstance(x, CharRange) else x.data
                            for x in self.chars))

class Repetition(Node):
    def __init__(self, t, start=None, parsed_start=None, data=None):
        super(Repetition, self).__init__(t, start, parsed_start, data)
        self.min = None
        self.max = None
        self.greedy = None

    def close(self, pos, parsed_pos, data):
        super(Repetition, self).close(pos, parsed_pos, data)

        if '*' in self.end_data:
            self.min, self.max = (0, None)
        elif '+' in self.end_data:
            self.min, self.max = (1, None)
        elif self.end_data[0] == '?':
            self.min, self.max = (0, 1)
        else:
            t = self.end_data.strip('{}?') # strip curlies
            if ',' in t:
                a, b = t.split(',')
                if a:
                    self.min = int(a)
                else:
                    self.min = 0
                if b:
                    self.max = int(b)
                else:
                    self.max = None
            else:
                self.min, self.max = (int(t), int(t))

        self.greedy = not (self.end_data.endswith('?') and self.end_data != '?')


class VerboseRegexTryAgain(Exception): pass


class BaseRegex(object):

    tokens = {
        'root': [
            (r'\|', Other.Alternate),
            (r'\(\?[iLmsux]+\)', Other.Directive),
            (r'\(\?:', Other.Open.NonCapturing),
            (r'(\(\?P<)(.*?)(>)', Other.Open.NamedCapturing),
            (r'\(\?=', Other.Open.Lookahead),
            (r'\(\?!', Other.Open.NegativeLookahead),
            (r'\(\?<!', Other.Open.NegativeLookbehind),
            (r'\(\?<', Other.Open.Lookbehind),
            (r'(\(\?P=)(\w+)(?=\))', Other.Open.ExistsNamed),
            (r'\(\?\(\d+\)', Other.Open.Exists),
            (r'\(\?#.*?\)', Other.Comment),
            (r'\(', Other.Open.Capturing),
            (r'\)', Other.CloseParen),
            (r'\[', Other.CharClass, 'charclass_start'),
            (r'\\[1-9][0-9]?', Other.Backref),
            include('only_in_verbose'),
            include('suspicious'),
            include('meta'),
            include('simpleliteral'),
            (r'[^\\()|\[\]]+', Other.Literals), # TODO
        ],
        'suspicious': [
            # misdone backreferences, tabs, newlines, and bel
            (r'[\x00-\x08\x0a\x0d]', Other.Suspicious),
        ],
        'charclass_start': [
            (r'\^', Other.NegateCharclass, 'charclass_squarebracket_special'),
            (r'', Other.Continue, 'charclass_squarebracket_special'),
        ],
        'charclass_squarebracket_special': [
            (r'\]', Other.Literal.CloseCharClass, 'charclass_rest'),
            (r'', Other.Continue, 'charclass_rest'),
        ],
        'charclass_rest': [
            (r'\]', Other.CloseCharClass, '#pop:3'),
            (r'\\-', Other.EscapedDash),
            (r'[\-^]', Other.Special),
            include('simpleliteral'),
            (r'\\.', Other.Suspicious),
        ],
        'meta': [
            (r'\.', Other.Dot),
            (r'\^', Other.Anchor.Beginning),
            (r'\$', Other.Anchor.End),
            (r'\\b', Other.Anchor.WordBoundary),
            (r'\\A', Other.Anchor.BeginningOfString),
            (r'\\Z', Other.Anchor.EndOfString),
            (r'\*\?', Other.Repetition.NongreedyStar),
            (r'\*', Other.Repetition.Star),
            (r'\+\?', Other.Repetition.NongreedyPlus),
            (r'\+', Other.Repetition.Plus),
            (r'\?\?', Other.Repetition.NongreedyQuestion),
            (r'\?', Other.Repetition.Question),
            (r'\{\d+,(?:\d+)?\}\??', Other.Repetition.Curly),
            (r'\{,?\d+\}\??', Other.Repetition.Curly),

        ],
        'simpleliteral': [
            (r'[^\\^-]', Other.Literal),
            (r'\\0[0-7]{0,3}', Other.Literal.Oct), # \0 is legal
            (r'\\x[0-9a-fA-F]{2}', Other.Literal.Hex),
            (r'\\[\[\]]', Other.Literal.Bracket),
            (r'\\[()]', Other.Literal.Paren),
            (r'\\t', Other.Tab),
            (r'\\n', Other.Newline),
            (r'\\\.', Other.Literal.Dot),
            (r'\\\\', Other.Literal.Backslash),
            (r'\\\*', Other.Literal.Star),
            (r'\\\+', Other.Literal.Plus),
            (r'\\\|', Other.Literal.Alternation),
            (r'\\\^', Other.Literal.Caret),
            (r'\\\$', Other.Literal.Dollar),
            (r'\\\?', Other.Literal.Question),
            (r'\\[{}]', Other.Literal.Curly),
            (r'\\\'', Other.Suspicious.Squo),
            (r'\\\"', Other.Suspicious.Dquo),
            (r'\\[sSwWdD]', Other.BuiltinCharclass),
            (r'\\.', Other.Suspicious), # Other unnecessary escapes
        ],
        'only_in_verbose': [],
    }

    @classmethod
    def get_parse_tree(cls, s, flags=0):
        if flags & re.VERBOSE:
            return VerboseRegex._get_parse_tree(s, flags)
        else:
            try:
                tree = cls._get_parse_tree(s, flags)
            except VerboseRegexTryAgain:
                #print "Using verbose mode"
                tree = VerboseRegex._get_parse_tree(s, flags | re.VERBOSE)
            return tree

    @classmethod
    def _get_parse_tree(cls, s, flags):
        n = RootNode(t=PROGRESSION, data='', start=0, parsed_start=0, raw=s,
                     flags=flags)
        verbose = flags & re.VERBOSE

        open_stack = [n]
        verbose_offset = 0
        #print "Using class", cls, cls.tokens['only_in_verbose']

        # these two need default values because normally they would be set in
        # the last iteration of the loop, but this doesn't happen for empty
        # string.
        j = 0
        data = ''

        # i, j are the raw position and parsed position, respectively.
        for i, ttype, data in cls().get_tokens_unprocessed(s):  # pylint: disable-msg=E1101
            #print i, ttype, data
            if not data: continue  # HACK for '' match for [][]

            if not verbose and ttype in Other.Directive and 'x' in data:
                raise VerboseRegexTryAgain()

            if ttype in Other.Verbose:
                #print "Found verbose token"
                verbose_offset += len(data)
                continue
            else:
                j = i - verbose_offset

            if ttype in Other.Open:
                # stack depth ++
                n = Node(t=ttype, start=i, parsed_start=j, data=data)
                open_stack.append(n)
            elif ttype is Other.CharClass and open_stack[-1].type is not Other.CharClass:
                n = CharClass(t=ttype, start=i, parsed_start=j)
                open_stack.append(n)
            elif ttype in (Other.CloseParen, Other.CloseCharClass):
                # stack depth -- or -= 2
                if open_stack[-2].type is ALTERNATION:
                    open_stack[-1].close(i, j, '')
                    open_stack[-2].add_child(open_stack[-1])
                    open_stack.pop()
                    open_stack[-1].close(i, j, '')
                    open_stack[-2].add_child(open_stack[-1])
                    open_stack.pop()
                assert (open_stack[-1].type in Other.Open or
                        open_stack[-1].type in Other.CharClass)
                open_stack[-1].close(i, j, data)
                open_stack[-2].add_child(open_stack[-1])
                open_stack.pop()
            elif ttype is Other.Alternate:
                # stack depth same, or +=2
                if len(open_stack) < 2 or open_stack[-2].type is not ALTERNATION:
                    # Create new alternation, push 2
                    #print s, open_stack[-1]
                    start = open_stack[-1].start + len(open_stack[-1].data)
                    parsed_start = open_stack[-1].parsed_start + len(open_stack[-1].data)
                    n = Node(t=ALTERNATION, start=start, parsed_start=parsed_start)
                    p = Node(t=PROGRESSION, start=start, parsed_start=parsed_start)
                    for c in open_stack[-1].children:
                        p.add_child(c) # sets parent
                    del open_stack[-1].children[:]
                    open_stack.append(n)
                    p.close(i, j, "")
                    n.add_child(p)
                    p2 = Node(t=PROGRESSION, start=i+len(data), parsed_start=j+len(data))
                    open_stack.append(p2)
                else:
                    # close & swap, replicating close a bit
                    open_stack[-1].close(i, j, "") # progression
                    open_stack[-2].add_child(open_stack[-1])
                    open_stack[-1] = Node(t=PROGRESSION, start=i+len(data),
                                          parsed_start=j+len(data))
            elif ttype in Other.Repetition:
                c = open_stack[-1].children.pop()
                n = Repetition(t=ttype, data='', start=c.start,
                               parsed_start=c.parsed_start)
                n.add_child(c)
                n.close(i, j, data)
                open_stack[-1].add_child(n)
            else:
                # stack depth same
                n = Node(t=ttype, data=data, start=i, parsed_start=j)
                open_stack[-1].add_child(n)

        # don't pop here
        if len(open_stack) == 3 and open_stack[-2].type is ALTERNATION:
            # TODO len(data) might fail if the regex is empty string, so
            # default above...
            open_stack[-1].close(len(s), j+len(data), '')
            open_stack[-2].add_child(open_stack[-1])
            open_stack[-2].close(len(s), j+len(data), '')
            open_stack[-3].add_child(open_stack[-2])
            open_stack.pop()
            open_stack.pop()

        open_stack[0].close(len(s), j+len(data), '')
        assert len(open_stack) == 1, s + repr(open_stack)
        return open_stack[0]


class Regex(BaseRegex, RegexLexer):
    r"""
    This is a RegexLexer that parses python regex syntax.  The included helper
    functions will turn its token stream into a psuedo-syntax-tree.  The goal
    is to recognize strange, error-prone regex constructs and provide warnings
    to users.  Examples of the sort of error-prone things:

    bad     good
    [][] = [\[\]]
    [A-z] = [\x41-\x7a] or [A-Za-z]
    '\n' = r'\n' or '\\n'

    Just because I wrote a parser in Pygments doesn't mean that it's generally
    a good idea.  Remember that.  This started because I need to run a regex
    lint against the Python strings, before the regex simplifier gets its hands
    on it.  This is fairly close to the best possible, but doesn't catch a few
    things (mainly in non-raw strings).
    """
    name = 'regex'
    mimetypes = ['text/x-regex']
    filenames = ['*.regex'] # fake
    flags = 0 # not multiline


class VerboseRegex(BaseRegex, RegexLexer):
    name = 'verbose_regex'
    mimetypes = ['text/x-verbose-regex']
    filenames = ['*.regex'] # fake
    flags = 0 # not multiline

    tokens = dict(BaseRegex.tokens)
    tokens['only_in_verbose'] = [
        (r'\s+', Other.Verbose.Whitespace),
        (r'#.*', Other.Verbose.Comment),
    ]


def charclass(c):
    if 'A' <= c <= 'Z':
        return 'upper'
    elif 'a' <= c <= 'z':
        return 'lower'
    else:
        return 'other'

def parser_main(args):
    if not args:
        regex = r'(foo|bar)|[ba]z'
    else:
        regex = args[0]

    r = Regex()
    #for x in r.get_tokens_unprocessed(regex):
    #    print x

    tree = r.get_parse_tree(regex)
    print '\n'.join(fmttree(tree))

if __name__ == '__main__':
    parser_main(sys.argv[1:])



# Copyright 2011-2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import sys
import logging

from pygments.token import Token

from regexlint.parser import *
from regexlint.util import *

def check_no_nulls(reg, errs):
    num = '101'
    level = logging.ERROR
    msg = 'Null characters not allowed (python docs)'
    pos = reg.raw.find('\x00')
    if pos != -1:
        errs.append((num, level, pos, msg))

def check_no_bels(reg, errs):
    num = '110'
    level = logging.ERROR
    msg = 'You probably don\'t want a backspace. Use another backslash, raw string, or use \\x08 instead)'
    pos = reg.raw.find('\b')
    if pos != -1:
        errs.append((num, level, pos, msg))

def check_no_newlines(reg, errs):
    num = '102'
    level = logging.ERROR
    msg = 'Newline characters not allowed (java compat, use a rawstring)'

    # Ignore re.VERBOSE modes for now.  I'm not sure how they fit in with
    # Java.
    if reg.flags & re.VERBOSE:
        return

    pos = reg.raw.find('\n')
    if pos != -1:
        errs.append((num, level, pos, msg))

def check_no_empty_alternations(reg, errs):
    num = '103'
    level = logging.ERROR
    msg = 'Empty string allowed in alternation starting at position %d, use ?'
    for n in find_all_by_type(reg, Other.Progression):
        if (not n.children and n.parent() and
            n.parent().type is Other.Alternation):
            errs.append((num, level, n.start or 0, msg % (n.start or 0)))

def check_charclass_homogeneous_ranges(reg, errs):
    num = '104'
    level = logging.ERROR
    msg = 'Range in character class is not homogeneous near position %d'
    msg2 = 'Range in character class goes backwards near position %d'
    for c in find_all_by_type(reg, Other.CharClass):
        for p in c.chars:
            if isinstance(p, CharRange):
                if (p.a.type is Other.Literal.Hex and
                    p.b.type is Other.Literal.Hex):
                    pass # hex notation for both sides ok to skip this check
                elif p.a.type is Other.Literal and p.b.type is Other.Literal:
                    # should be single character data, can compare
                    assert len(p.a.data) == 1
                    assert len(p.b.data) == 1
                    if charclass(p.a.data) != charclass(p.b.data):
                        errs.append((num, level, p.a.start, msg % p.a.start))
                    # only positive ranges are allowed.
                    if ord(p.a.data) >= ord(p.b.data):
                        errs.append((num, level, p.a.start, msg2 % p.a.start))
                elif p.a.type not in Other.Literal and p.b.type not in Other.Literal:
                    # punctuation range?
                    errs.append((num, level, p.a.start, msg % p.a.start))
                elif (p.a.type in (Other.Literal.Unicode,
                                   Other.Literal.LongUnicode) and
                      p.b.type in (Other.Literal.Unicode,
                                   Other.Literal.LongUnicode)):
                    pass # ok
                else:
                    # strange range.
                    errs.append((num, level, p.a.start, msg % p.a.start))

def check_prefix_ordering(reg, errs):
    """
    Checks for things of the form a|ab, which should be ab|a due to python
    quirks.
    """
    num = '105'
    level = logging.ERROR
    msg = 'Potential out of order alternation between %r and %r'
    for n in find_all_by_type(reg, Other.Alternation):
        run_checks = True
        for i in between(n, None):
            # TODO this heuristic is easy to game
            if i.type in Other.Anchor or i.type in Other.Open or width(i.type):
                run_checks = False
                break
        if not run_checks:
            continue

        prev = None
        for i in n.children:
            assert i.type is Other.Progression
            #print i, reg.raw[n.start:n.end]
            if not all(x.type in Other.Literal or
                       x.type in Other.Literals or
                       x.type in Other.Newline or
                       x.type in Other.Suspicious
                       for x in i.children):
                #print "Can't check", i
                return
            t = ''.join(x.data for x in i.children)
            #print "Check", repr(t), repr(prev)
            if prev is not None and t.startswith(prev):
                errs.append((num, level, i.start, msg % (prev, t)))
                break
            prev = t

def check_no_python_named_capture_groups(reg, errs):
    num = '106'
    level = logging.ERROR
    msg = 'Python named capture group'
    for n in find_all_by_type(reg, Other.Open.NamedCapturing):
        errs.append((num, level, n.start, msg))
        break

def bygroups_check_toknum(reg, errs, desired_groups):
    num = '107'
    level = logging.ERROR
    msg = 'Wrong number of groups(%d) for bygroups(%d)'
    n = len(list(find_all_by_type(reg, Other.Open.Capturing)))
    desired_number = len(desired_groups)
    if n < desired_number:
        errs.append((num, level, 0, msg % (n, desired_number)))
    elif n > desired_number:
        # If there are nested groups anywhere but the last, check_overlap will
        # find them.  This checker doesn't look at start/stop positions.
        errs.append((num, logging.INFO, 0,
                     (msg % (n, desired_number)) + ' (extra groups)'))

def bygroups_check_overlap(reg, errs, desired_groups):
    num = '108'
    level = logging.ERROR
    msg = 'Nested capture group other than the final one using bygroups'
    msg2 = 'Gap in capture groups using bygroups'
    n = list(find_all_by_type(reg, Other.Open.Capturing))
    if not n:
        # bygroups_check_toknum should already complain about this case.
        return
    desired_number = len(desired_groups)
    # The order returned by find_all_by_type appears to be the same as python's
    # group numbers (matters most when nesting).
    prev_end = 0
    prev = None
    #print reg.raw, desired_number
    for idx, group in enumerate(n):
        if group.parent().type in Other.Repetition:
            group = group.parent()
        #print "Loop", i, group, group.start, group.end

        if group.parsed_start > prev_end:
            #print "Have prev"
            # This code allows a parent to be ok'd, and all children to be
            # ignored (without having to change between()'s code)
            j = find_bad_between(prev, group, has_width)
            if j:
                errs.append((num, level, j.start, msg2))
        elif group.parsed_start < prev_end:
            if idx >= desired_number:
                # This case is uninteresting -- bygroups ignores extra groups,
                # so it's possible to nest within the last group.
                errs.append((num, logging.INFO, group.start,
                             msg + ' (extra groups)'))
                group = prev
            else:
                # This is a nested group with the outer one prior to the last
                # one bygroups cares about
                #print "Boring", group.start, prev_end
                if desired_groups[idx] is not None:
                    errs.append((num, level, group.start, msg))
                group = prev

        prev_end = group.parsed_end
        prev = group

    if prev_end != reg.parsed_end:
        #print "End check", prev
        # This code allows a parent to be ok'd, and all children to be
        # ignored (without having to change between()'s code)
        j = find_bad_between(prev, None, has_width)
        if j:
            errs.append((num, level, j.start, msg))

def bygroups_check_no_capture_group_in_repetition(reg, errs, desired_groups):
    num = '109'
    level = logging.ERROR
    msg = 'Capture group should not be within a repetition when using bygroups'
    desired_number = len(desired_groups)
    for idx, capture in enumerate(find_all_by_type(reg, Other.Open.Capturing)):
        parent = capture.parent()
        while parent:
            # Question works in Pygments at the moment, but is subject to change.
            if (parent.type in Other.Repetition and
                parent.type is not Other.Repetition.Question):
                if idx >= desired_number:
                    errs.append((num, logging.INFO, capture.start, msg + ' (extra groups)'))
                elif desired_groups[idx] is not None:
                    errs.append((num, level, capture.start, msg))
            parent = parent.parent()


def check_no_consecutive_dots(reg, errs):
    num = '111'
    level = logging.WARNING
    msg = 'Consecutive dots, use .{2} if this is intentional'
    for x in find_all_by_type(reg, Other.Dot):
        n = x.next_no_children()
        if n and n.type is Other.Dot:
            errs.append((num, level, x.start, msg))
            break

def check_unicode_escapes(reg, errs):
    num = '112'
    level = logging.ERROR
    msg = 'Regex parser does not handle unicode, use ur"" string.'
    r_unicode = re.compile(r'\\[uU][0-9a-fA-F]|\\N{')
    for m in r_unicode.finditer(reg.raw):
        errs.append((num, level, m.start(), msg))

def check_bad_flags(reg, errs):
    num = '113'
    level = logging.WARNING
    msg = 'Manually set flag %r, but do not need it'

    directives = list(find_all_by_type(reg, Other.Directive))
    # TODO flag the correct directive
    flags = ''.join(d.data for d in directives)
    if not flags:
        return

    if 'x' in flags:
        # In order for x to matter, there must exist a node that has start !=
        # parsed_start.  The easiest place to find this is on the last one, since
        # both values should be nondecreasing.
        if reg.children[-1].parsed_end == len(reg.raw):
            errs.append((num, level, directives[0].start, msg % 'x'))

        # TODO See if there's bare whitespace
        pass

    if 'i' in flags:
        # See if there are a-zA-Z
        try:
            for char in find_all_by_type(reg, Other.Literal):
                if 'a' <= char.data <= 'z' or 'A' <= char.data <= 'Z':
                    raise Break()

            # This part only checks ranges, because the single characters were
            # already checked directly above.
            import string
            alpha = set(map(ord, string.letters))
            for cc in find_all_by_type(reg, Other.CharClass):
                for char in cc.chars:
                    if isinstance(char, CharRange):
                        this_range = set(range(char.codepoint_a, char.codepoint_b))
                        if this_range & alpha:
                            raise Break()
        except Break:
            pass
        else:
            errs.append((num, level, directives[0].start, msg % 'i'))

    if 's' in flags:
        # See if there are any dots.
        dots = list(find_all_by_type(reg, Other.Dot))
        if not dots:
            errs.append((num, level, directives[0].start, msg % 's'))

    if 'm' in flags:
        # Only ^$ differ in this mode.
        anchors = list(find_all_by_type(reg, (Other.Anchor.Beginning, Other.Anchor.End)))
        if not anchors:
            errs.append((num, level, directives[0].start, msg % 'm'))


def check_suspicious_anchors(reg, errs):
    num = '114'
    level = logging.WARNING
    msg = 'Suspicious use of anchors in alternation'

    for rep in find_all_by_type(reg, Other.Alternation):
        first = rep
        while first.children:
            first = first.children[0]

        last = rep
        while last.children:
            last = last.children[-1]

        if first.type in Other.Anchor and last.type in Other.Anchor:
            errs.append((num, level, first.start, msg))


def check_single_character_classes(reg, errs):
    num = '115'
    level = logging.INFO # harmless, for now
    msg = 'Only a single character in character class'

    for cc in find_all_by_type(reg, Other.CharClass):
        if (len(cc.chars) == 1 and
            not cc.negated and
            cc.parent().type not in Other.Repetition and
            (not isinstance(cc.chars[0], CharRange) or
             cc.chars[0].codepoint_a == cc.chars[0].codepoint_b)):
            errs.append((num, level, cc.start, msg))


def check_unicode_strings_actually_contain_unicode(reg, errs):
    num = '116'
    level = logging.WARNING
    msg = 'Unicode string, but contains only ascii'

    if isinstance(reg.raw, unicode):
        if max(ord(c) for c in reg.raw) < 128:
            errs.append((num, level, 0, msg))


def check_charclass_overlap(reg, errs):
    num = '117'
    level = logging.WARNING
    msg = 'Overlap in character class: %r'

    for cc in find_all_by_type(reg, Other.CharClass):
        if len(set(cc.matching_character_codes)) != len(cc.matching_character_codes):
            counts = {}
            for i in cc.matching_character_codes:
                counts.setdefault(i, 0)
                counts[i] += 1
            dupes = [chr(k) for k, v in counts.iteritems() if v > 1]
            errs.append((num, level, cc.start, msg % (dupes,)))

def check_charclass_case_insensitive_overlap(reg, errs):
    num = '122'
    level = logging.WARNING
    msg = 'Overlap due to case insensitive mode'

    directives = list(find_all_by_type(reg, Other.Directive))
    flags = ''.join(d.data for d in directives)
    if 'i' not in flags and not re.IGNORECASE & reg.flags:
        return

    def fold(i):
        if i >= 97 and i <= 122:
            return i - 32
        return i

    # TODO: This only finds the most obvious ones, like
    # (?i)[0-9a-fA-F], and doesn't do anything about non-ranges.
    for cc in find_all_by_type(reg, Other.CharClass):
        ranges = set()
        for c in cc.chars:
            if isinstance(c, CharRange):
                a = eval_char(c.a.data)
                b = eval_char(c.b.data)
                if (fold(a), fold(b)) in ranges:
                    errs.append((num, level, c.a.start, msg))
                ranges.add((fold(a), fold(b)))

COMMON_SINGLE_CHAR_CODES = map(ord, '()*+. ')

def check_charclass_len(reg, errs):
    num = '118'
    level = logging.WARNING
    msg = 'Superfluous character class when only one char'

    for cc in find_all_by_type(reg, Other.CharClass):
        if not cc.negated and len(cc.matching_character_codes) == 1:
            # Some people use [*] instead of \* -- allow this for now as an INFO
            if (cc.matching_character_codes[0] in COMMON_SINGLE_CHAR_CODES
                or cc.parent().type in Other.Repetition):
                errs.append((num, logging.INFO, cc.start, msg))
            else:
                errs.append((num, level, cc.start, msg))


def check_charclass_negation(reg, errs):
    num = '119'
    level = logging.WARNING
    msg = 'Instead of negating character class, flip case of builtin class'

    for cc in find_all_by_type(reg, Other.CharClass):
        if (cc.negated and len(cc.children) == 2 and
            cc.children[1].type in Other.BuiltinCharclass):
            errs.append((num, level, cc.start, msg))


def check_multiline_anchors(reg, errs):
    num = '120'
    level = logging.WARNING
    msg = 'Use of ^ or $ without multiline mode: use \\A or \\Z explicitly.'

    if reg.flags & re.M:
        return

    for anchor in find_all_by_type(reg, (Other.Anchor.Beginning,
                                         Other.Anchor.End)):
        errs.append((num, level, anchor.start, msg))


def check_wide_unicode(reg, errs):
    num = '121'
    level = logging.WARNING
    msg = 'Wide unicode causes problems in narrow builds'

    if isinstance(reg.raw, unicode):
        for lit in find_all_by_type(reg, Other.Literal):
            if len(lit.data) == 1 and ord(lit.data) > 65535:
                # entire codepoint, we're in a wide build
                if isinstance(lit.parent(), CharClass) or lit.parent().type in Other.Repetition:
                    errs.append((num, level, lit.start, msg))
            elif (len(lit.data) == 1 and 0xd800 <= ord(lit.data) <= 0xdbff and
                  isinstance(lit.parent(), CharClass)):
                # high surrogate byte, we're in a narrow build, does the wrong thing
                n = lit.next_no_children()
                if (n is not None and n.type is Other.Literal and len(n.data) == 1 and
                    0xdc00 <= ord(n.data) <= 0xdfff):
                    errs.append((num, level, lit.start, msg))
            elif (len(lit.data) == 1 and 0xdc00 <= ord(lit.data) <= 0xdfff and
                  lit.parent().type in Other.Repetition):
                # low surrogate byte, imagine HH LL +
                errs.append((num, level, lit.start, msg))
            # TODO figure out if there's a way to catch overly-verbose unicode
            # (needs to happen before string parsing)
            # TODO expand to more use of suspicious unicode


def manual_check_for_empty_string_match(reg, errs, raw_pat):
    # Note, things like '#pop' and 'next-state' get a pass on this, as
    # do callback functions, since they are mostly used for advanced
    # features in indent-dependent languages.
    if len(raw_pat) < 3 and isinstance(raw_pat[1], Token.__class__):
        regex = re.compile(raw_pat[0])
        # Either match on empty string, or empty string at the end of a word
        if regex.match('') or regex.match('a', 1):
            errs.append(('999', logging.ERROR, 0, 'Matches empty string'))
        #remove_error(errs, '103')


def run_all_checkers(regex, expected_groups=None):
    errs = []
    for k, f in globals().iteritems():
        if k.startswith('check_'):
            #print 'running', k
            try:
                f(regex, errs)
            except Exception, e:
                errs.append(('999', logging.ERROR, 0, "Checker %s encountered error parsing: %s" % (f, repr(e))))
        elif k.startswith('bygroups_check_') and expected_groups:
            try:
                f(regex, errs, expected_groups)
            except Exception, e:
                errs.append(('999', logging.ERROR, 0, "Checker %s encountered error parsing: %s" % (f, repr(e))))
    return errs

def main(args):
    if not args:
        regex = r'(foo|) [a-Mq-&]'
    else:
        regex = args[0]
    print run_all_checkers(Regex.get_parse_tree(regex))

if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/env python2
#
# Copyright 2011-2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import re
import logging
import itertools
import multiprocessing

from StringIO import StringIO  # can't pickle cStringIO

from pygments.lexer import RegexLexer, bygroups
from pygments.token import Token
from regexlint import Regex, run_all_checkers
from regexlint.checkers import manual_check_for_empty_string_match
from regexlint.indicator import find_offending_line, mark, find_substr_pos
from regexlint.util import consistent_repr, shorten

def import_mod(m):
    mod = __import__(m)
    return sys.modules[m]

def main(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    import optparse
    o = optparse.OptionParser()
    o.add_option('--min_level',
                 help='Min level to print (logging constant names like ERROR)',
                 default='WARNING')
    o.add_option('--output_file',
                 help='Output filename for analysis',
                 default=None)
    o.add_option('--no_parallel',
                 help='Run checks in a single thread',
                 default=True,
                 dest='parallel',
                 action='store_false')
    opts, args = o.parse_args(argv)

    min_level = getattr(logging, opts.min_level)
    if opts.output_file:
        output_stream = file(opts.output_file, 'wb')
    else:
        output_stream = sys.stdout

    # currently just a list of module names.
    lexers_to_check = []
    for module in args:
        if ':' in module:
            module, cls = module.split(':')
        else:
            cls = None
        mod = import_mod(module)
        lexers_to_check.append(StringIO("Module %s\n" % module))
        if cls:
            lexers = [cls]
        else:
            if hasattr(mod, '__all__'):
                lexers = mod.__all__
            else:
                lexers = mod.__dict__.keys()

        for k in lexers:
            v = getattr(mod, k)
            if hasattr(v, '__bases__') and issubclass(v, RegexLexer) and v.tokens:
                lexers_to_check.append((k, v, mod.__file__, min_level,
                                        StringIO()))

    if opts.parallel:
        pool = multiprocessing.Pool()
    else:
        pool = itertools

    for result in pool.imap(check_lexer_map, lexers_to_check):
        result.seek(0, 0)
        output_stream.write(result.read())


def remove_error(errs, *nums):
    for i in range(len(errs)-1, -1, -1):
        if errs[i][0] in nums:
            del errs[i]

def check_lexer_map(args):
    if isinstance(args, StringIO):
        return args
    return check_lexer(*args)

def check_lexer(lexer_name, cls, mod_path, min_level, output_stream=sys.stdout):
    #print lexer_name
    #print cls().tokens
    has_errors = False

    bygroups_callback = bygroups(1).func_code
    for state, pats in cls().tokens.iteritems():
        if not isinstance(pats, list):
            # This is for Inform7Lexer
            print >>output_stream, lexer_name, 'WEIRD'
            return output_stream

        for i, pat in enumerate(pats):
            #print repr(pat[0])
            try:
                reg = Regex.get_parse_tree(pat[0], cls.flags)
            except TypeError:
                # Doesn't support _inherit yet.
                continue
            except:
                print >>output_stream, pat[0], cls
                raise
            # Special problem: display an error if count of args to
            # bygroups(...) doesn't match the number of capture groups
            if callable(pat[1]) and pat[1].func_code is bygroups_callback:
                by_groups = pat[1].func_closure[0].cell_contents
            else:
                by_groups = None

            errs = run_all_checkers(reg, by_groups)
            # Special case for empty string, since it needs action.
            manual_check_for_empty_string_match(reg, errs, pat)

            errs.sort(key=lambda k: (k[1], k[0]))
            if errs:
                #print "Errors in", lexer_name, state, "pattern", i
                for num, severity, pos1, text in errs:
                    if severity < min_level: continue

                    # Only set this if we're going to output something --
                    # otherwise the [Lexer] OK won't print
                    has_errors = True

                    foo = find_offending_line(mod_path, lexer_name, state, i,
                                              pos1)
                    if foo:
                        line = 'L' + str(foo[0])
                    else:
                        line = 'pat#' + str(i+1)
                    print >>output_stream, '%s%s:%s:%s:%s: %s' % (
                        logging.getLevelName(severity)[0], num,
                        lexer_name, state, line, text)
                    if foo:
                        mark(*(foo + (output_stream,)))
                    else:
                        # Substract one for closing quote
                        start = len(consistent_repr(pat[0][:pos1])) - 1
                        end = len(consistent_repr(pat[0][:pos1+1])) - 1
                        if start == end:
                            # This handles the case where pos1 points to the end
                            # of the string. Regex "|" with pos1 = 1.
                            end += 1
                        assert end > start
                        text, start, end = shorten(repr(pat[0]), start, end)
                        mark(-1, start, end, text, output_stream)
    if not has_errors:
        print >>output_stream, lexer_name, "OK"

    return output_stream


if __name__ == '__main__':
    main()



# Copyright 2011-2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import with_statement
from pygments.token import Other


def find_all(first, second=None):
    """Finds all descendants (inorder) of first, including itself.  If second
    is provided, stops when it is reached."""
    regex = first
    while regex and regex is not second:
        yield regex
        regex = regex.next()


def find_all_by_type(regex_root, t):
    for regex in find_all(regex_root):
        if regex.type in t:
            yield regex


def between(first, second):
    """Yields all nodes between first and second, not including either
    endpoint.  The special first value None means to start at the beginning,
    not including the root.  The special second value None means to go until
    the end, but either first or second must be provided."""
    if first is None:
        first = second
        while first.parent():
            first = first.parent()
        first = first.children[0]
    else:
        first = first.next_no_children()

    it = find_all(first, second)
    for i in it:
        yield i


def find_bad_between(first, second, fn):
    """Finds the first node in between(first, second) where fn returns True.
    If fn returns None, the node will be descended, False, it will be ignored.
    The default return value is None."""

    # Rather than a complex, fast solution, this simply walks all nodes, and
    # ignores the ones that already have a whitelisted parent.
    good_obj = None
    for j in between(first, second):
        #print "Intermediate", j, j.type
        if good_obj and j.is_descentant_of(good_obj):
            pass
        else:
            v = fn(j)
            if v == True:
                return j
            elif v == False:
                good_obj = j
            # else keep going


def has_width(node):
    """Helper for use with find_bad_between."""
    # returns True/False/None
    return width(node.type)


def fmttree(t):
    """Returns a list of lines containing a pretty-printed tree.
    The tree is any regex node, passed in as t.  Indent is two spaces.
    """
    if not hasattr(t, 'children'):
        return [repr(t)]

    r = []
    r.append('<%s type=%r data=%r>' % (t.__class__.__name__, t.type, t.data))
    for c in t.children:
        r.extend('  ' + f for f in fmttree(c))
    return r


def width(tok):
    """Returns whether the given token type might consume characters."""
    if (tok in (Other.Directive,
                Other.Open.Lookahead, Other.Open.NegativeLookahead,
                Other.Open.NegativeLookbehind, Other.Open.Lookbehind,
                Other.Comment) or (tok in Other.Anchor)):
        return False
    elif (tok in Other.Open or tok in Other.Alternation or tok in
          Other.Progression or tok in Other.Repetition):
        return None # unsure if it has width, must descend
    else:
        return True

def eval_char(c):
    """Returns the character code of the string s, which may contain
    escapes."""
    if len(c) == 1:
        return ord(c)
    elif c[-1] == "'":
        return ord("'")
    elif c[0] == '\\' and c[1] not in 'abtrnvfx01234567\\':
        c = c[1:] # unnecessary backslash?

    #print repr(c)
    if len(c) == 1 and ord(c) >= 128:
        return c

    try:
        return ord(eval("'%s'" % c))
    except TypeError:
        print "C is", repr(c), len(c)
        raise
    # TODO any other cases?

class Break(Exception):
    pass

def consistent_repr(s):
    """Returns a string that represents repr(s), but without the logic that
    switches between single and double quotes."""
    special = {
        '\n': '\\n',
        '\t': '\\t',
        '\\': '\\\\',
        '\'': '\\\'',
    }
    rep = ['\'']
    if isinstance(s, unicode):
        rep.insert(0, 'u')
    for char in s:
        if char in special:
            rep.append(special[char])
        elif isinstance(s, unicode) and ord(char) > 0xFFFF:
            rep.append('\\U%08x' % ord(char))
        elif isinstance(s, unicode) and ord(char) > 126:
            rep.append('\\u%04x' % ord(char))
        elif ord(char) < 32 or ord(char) > 126:
            rep.append('\\x%02x' % ord(char))
        else:
            rep.append(char)
    rep.append('\'')
    return ''.join(rep)

def shorten(text, start, end):
    """Returns a substring of text, so that its length is < 80 or so."""
    if len(text) < 76:
        return (text, start, end)

    start_cut = max(0, start - 36)
    end_cut = min(len(text), start + 36)
    cut_text = text[start_cut:end_cut]
    start -= start_cut
    end -= start_cut
    if start_cut != 0:
        cut_text = '...' + cut_text
        start += 3
        end += 3
    if end_cut != len(text):
        cut_text += '...'
    return (cut_text, start, end)

def get_module_text(mod):
    if '\n' in mod:
        mod_text = mod
    else:
        if mod.endswith('.pyc') or mod.endswith('.pyo'):
            mod = mod[:-1]
        with file(mod, 'r') as f:
            mod_text = f.read()
    return mod_text

def rindex(a, x):
    for i in range(len(a)-1, -1, -1):
        if a[i] == x:
            return i
    raise ValueError("Not found")




# Copyright 2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from regexlint.util import rindex

strp = {
    '': re.compile(r'\\(?:[\\abfnrtv"\']|\n|x[a-fA-F0-9]{2}|[0-7]{1,3})|'
                   r'[\w\W]'),
    'u': re.compile(r'\\(?:[\\abfnrtv"\']|\n|N{.*?}|u[a-fA-F0-9]{4}|'
                    r'U[a-fA-F0-9]{8}|x[a-fA-F0-9]{2}|[0-7]{1,3})|[\w\W]'),
    'r': re.compile(r'[\w\W]'),
    'ur': re.compile(r'\\(?:\\|u[a-fA-F0-9]{4}|U[a-fA-F0-9]{8})|[\w\W]'),
}

def find_substr_pos(s, target):
    if s[-3:] in ('"""', "'''"):
        end_quote = s[-3:]
    else:
        end_quote = s[-1]
    p = s.find(end_quote)
    mods = s[:p]
    body = s[p+len(end_quote):-len(end_quote)]

    chars = strp[mods].findall(body)
    if 'r' not in mods:
        # hack to support a zero-width escape sequence -- escaped newline
        for i in range(len(chars)):
            if chars[i] == '\\\n':
                target += 1
            if i >= target: break

    if target >= len(chars) or target < 0:
        raise ValueError("Impossible, out of bounds")

    l = 0
    q = p+len(end_quote)+sum(map(len, chars[:target]))

    # only for triplequoted strings
    if '\n' in chars[:target]:
        #print "GOT", target, chars[:target]
        l = chars[:target].count('\n')
        _ = rindex(chars[:target], '\n') + 1
        #print "_", _
        q = sum(map(len, chars[_:target]))
        #print "q", l, q, q+len(chars[target])

    #print q
    return (l, q, q+len(chars[target]))




# Copyright 2011-2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script is several layered hacks to be able to point out clang-style errors
for a specific line of Python code (likely in a string literal).  Do not take
as an example of well-written Python.
"""
import re

from pygments.lexers.agile import PythonLexer
from pygments.token import Punctuation, Name, Text, String

from regexlint.indicator_substr import find_substr_pos
from regexlint.util import get_module_text

R_CLASSEXTRACT = re.compile(r'class (\w+).*:\n([\w\W]*?)(?=^class|\Z)', re.MULTILINE)

open_braces = '[{('
close_braces = ']})'


def find_offending_line(mod, clsname, state, idx, pos):
    """
    Returns a tuple of (lineno, charpos_start, charpos_end, line_content)
    """
    mod_text = get_module_text(mod)

    # skip as far as possible using regular expressions, then start lexing.
    for m in R_CLASSEXTRACT.finditer(mod_text):
        #print "Got class", m.group(1), '...' + repr(m.group(0)[-40:])
        if m.group(1) == clsname: break
    else:
        return None
        #raise ValueError("Can't find class %r" % (clsname,))

    def match_brace(brace):
        target = close_braces[open_braces.index(brace)]
        #print "matching brace", brace, target, it
        for x in it:
            #print x
            if x[-2] in Punctuation and x[-1] == target:
                #print "  found"
                return True
            elif x[-2] in Punctuation and x[-1] in open_braces:
                # begin again!
                if not match_brace(x[-1]):
                    #print "  fail2"
                    return False
        #print "  inner fail"
        return False

    def until(t, match_braces=False):
        #print "until", t
        for x in it:
            if t == x[-1]:
                #print "  found", repr(x[-1])
                return
            elif match_braces and x[-1] in open_braces:
                match_brace(x[-1])

    level = 0
    tuple_idx = 0
    string_pos = 0

    def amal(i):
        si = None
        pt = None
        pd = None
        for idx, tok, data in i:
            if tok in String: tok = String
            if tok == pt == String:
                pd += data
            else:
                if pt:
                    yield si, pt, pd
                pt = tok
                pd = data
                si = idx
        if pt:
            yield si, pt, pd

    def filt(i):
        line = 1 + mod_text[:m.start()].count('\n')
        col = 0
        for _, b, c in i:
            #print "got", b, repr(c)
            yield line, col, b, c
            line += c.count('\n')
            if '\n' in c:
                col = len(c) - c.rindex('\n') - 1
            else:
                col += len(c)

    it = filt(amal(PythonLexer().get_tokens_unprocessed(m.group(0))))

    for x, y, ttyp, text in it:
        #print "Loop", level, ttyp, repr(text)
        if level == 0 and ttyp is Name:
            if text == 'tokens':
                until('=')
                until('{')
                level = 1
        elif level == 1 and ttyp in String:
            #print "key", text
            key = eval(text, {}, {})

            # next is expected to be the colon.
            it.next()
            # next is either a left brace, or maybe a function call
            t = ''
            try:
                while not t.strip():
                    _, _, _, t = it.next()
            except StopIteration:
                return None
            # t is now the first token of the value, either '[' or 'SomeFunc'

            if key != state:
                if t == '[':
                    match_brace('[')
                else:
                    until(',', match_braces=True)
            else:
                level = 2
                if t != '[':
                    return None
        elif level == 2:
            if text == '(':
                # open a tuple
                level = 3
            elif text == ')':
                level = 1 # honestly this should be able to just return
                #print "too late", idx, tuple_idx
                return
        elif level == 3:
            #print "  idx", tuple_idx
            if text == ')':
                level = 2
                tuple_idx += 1
            elif text == '(':
                match_brace('(')
            elif tuple_idx == idx and ttyp in String:
                # this might be it!
                s = eval(text, {}, {})
                #print "maybe", string_pos, pos, (string_pos+len(s))
                if string_pos <= pos < (string_pos+len(s)):
                    # need to point in here
                    (dx, d1, d2) = find_substr_pos(text, pos - string_pos)
                    if dx == 0:
                        d1 += y
                        d2 += y
                    return (x+dx, d1, d2, mod_text.splitlines()[x+dx-1])
                else:
                    string_pos += len(s)
            elif tuple_idx == idx and ttyp in Name:
                # If they're concatenating strings with vars, ignore.
                break




from setuptools import setup, find_packages
import os.path

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()

version = '0.9.2'

setup(name='regexlint',
      version=version,
      description='Linter for (Pygments) regular expressions',
      long_description=README,
      classifiers = [
        'License :: OSI Approved :: Apache Software License',
        'Intended Audience :: Developers',
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent', # I hope
        'Topic :: Software Development :: Quality Assurance', # Closest to linting I see
      ],
      keywords='',
      author='Tim Hatch',
      author_email='tim@timhatch.com',
      url='https://github.com/thatch/regexlint',
      license='Apache',
      packages=find_packages('.'),
      install_requires=['Pygments'],
      entry_points={
        'console_scripts': [
            'regexlint=regexlint.cmdline:main',
        ],
      },
      test_suite='nose.collector',
      tests_require=['nose'],
)



# Copyright 2011-2012 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This uses the python ast module to find the positions of strings in the source
file.  Due to what appear to be bugs in the ast module when a string spans
lines, this is not the default.
"""

import re
import ast

from regexlint.indicator_substr import find_substr_pos
from regexlint.util import get_module_text

try:
    enumerate((), 1)
except TypeError:
    # Py25 compat
    def enumerate(l, start=None):
        i = start or 0
        for x in l:
            yield i, x
            i += 1

def find_offending_line(mod, clsname, state, idx, pos):
    """
    Returns a tuple of (lineno, charpos_start, charpos_end, line_content)
    """
    mod_text = get_module_text(mod)

    tree = ast.parse(mod_text)
    klass = None
    for item in ast.walk(tree):
        if isinstance(item, ast.ClassDef):
            if clsname == item.name:
                klass = item
                break
    if not klass:
        return
    for item in ast.walk(klass):
        if not isinstance(item, ast.Dict):
            continue
        stateValue = None
        for i, key in enumerate(item.keys):
            if isinstance(key, ast.Str) and key.s == state:
                stateValue = item.values[i]
        if stateValue == None:
            continue
        if not isinstance(stateValue, ast.List):
            continue
        if not idx < len(stateValue.elts):
            continue
        idxTuple = stateValue.elts[idx]
        if not isinstance(idxTuple, ast.Tuple):
            continue
        if len(idxTuple.elts) < 2 or not isinstance(idxTuple.elts[0], ast.Str):
            continue
        lines = []
        startline = idxTuple.elts[0].lineno - 1
        startchar = idxTuple.elts[0].col_offset
        stopline = idxTuple.elts[1].lineno
        stopchar = idxTuple.elts[1].col_offset

        #print "Block"
        #print idxTuple.col_offset, idxTuple.elts[0].col_offset, idxTuple.elts[1].col_offset
        #print idxTuple.lineno, idxTuple.elts[0].lineno, idxTuple.elts[1].lineno
        #print mod_text.splitlines()[idxTuple.lineno-1]

        # HACK HACK HACK
        if idxTuple.elts[0].col_offset == -1:
            # When col_offset==-1 then the string is split across multiple
            # lines. This generally means it's a triplequoted string, which
            # can be located with a dumb heuristic.
            _lines = mod_text.splitlines()
            if '"""' in _lines[startline]:
                target = '"""'
            elif "'''" in _lines[startline]:
                target = "'''"
            else:
                #print "Cannot locate beginning string"
                return None

            for i in range(startline-1, -1, -1):
                if target in _lines[i]:
                    #print "Adjusted start line from", startline, "to", i
                    startline = i
                    startchar = _lines[i].rindex(target) - 2 # for 'ur'
                    break

        # END HACK

        for i, line in enumerate(mod_text.splitlines()[startline:], startline):
            if i == stopline:
                line = line[:stopchar]
            if i == startline:
                line = line[startchar:]
            lines.append(line)
            if i >= stopline:
                break
        rawstr = "\n".join(lines)

        strRe = re.compile("[uU]?[rR]?(?:"
            "'''(?:[^\\\\]|\\\\.)*?'''|"
            '"""(?:[^\\\\]|\\\\.)*?"""|'
            "'(?:[^\\\\]|\\\\.)*?'|"
            '"(?:[^\\\\]|\\\\.)*?"'
            ")", re.DOTALL)
        #print "rawstr:", repr(rawstr)
        for match in strRe.finditer(rawstr):
            #print "match:", repr(match.group(0))
            strInst = match.group(0)
            try:
                (dx, d1, d2) = find_substr_pos(strInst, pos)
            except ValueError:
                pos -= len(eval(strInst))
                continue
            before_match = rawstr[:match.start(0)]
            match_lineno_in_rawstr = before_match.count("\n")
            lineno = startline + 1 + match_lineno_in_rawstr
            col_offset = match.start(0) - (before_match.rfind("\n") + 1)
            if match_lineno_in_rawstr == 0:
                col_offset += startchar
            if dx == 0:
                d1 += col_offset
                d2 += col_offset
            return (lineno+dx, d1, d2,
                    mod_text.splitlines()[lineno+dx-1])
